@@ -1,26 +1,50 @@
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
+/* 
+ * Copyright (C) 2002-2004 Edd Dumbill <edd@usefulinc.com>
+ * Copyright (C) 2005-2007 Bastien Nocera <hadess@hadess.net>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ *
+ */
 
 #include <stdlib.h>
+#include <string.h>
 #include <glib.h>
+#include <glib/gi18n.h>
 
 #include "app.h"
 
-#define POLL_TIMEOUT 50
-
-static gchar *
+static gboolean
 set_connection_device (MyApp *app)
 {
-	gint ctype = gconf_client_get_int (app->client,
-			CONFBASE"/connection_type", NULL);
-	gchar *bdaddr;
-	gchar *dev = NULL;
+	int ctype;
+	char *dev = NULL;
+	gboolean changed;
 
-	bdaddr = gconf_client_get_string (app->client,
-			CONFBASE"/bluetooth_addr", NULL);
+	ctype = gconf_client_get_int (app->client,
+				      CONFBASE"/connection_type", NULL);
+	changed = FALSE;
 
 	switch (ctype) {
 		case CONNECTION_BLUETOOTH:
-			if (bdaddr && strlen (bdaddr) == 17) {
-				dev = g_strdup (bdaddr);
+			dev = gconf_client_get_string (app->client,
+						       CONFBASE"/bluetooth_addr",
+						       NULL);
+			if (!dev || strlen (dev) != 17) {
+				g_free (dev);
+				dev = NULL;
 			}
 			break;
 		case CONNECTION_SERIAL1:
@@ -38,25 +62,22 @@ set_connection_device (MyApp *app)
 			break;
 	}
 
-	if (bdaddr)
-		g_free (bdaddr);
+	if (dev == NULL && app->devname != NULL)
+		changed = TRUE;
+	else if (dev != NULL && app->devname == NULL)
+		changed = TRUE;
+	else if (dev != NULL && app->devname != NULL && strcmp (dev, app->devname) != 0)
+		changed = TRUE;
 
-	if (dev) {
-		if (app->devname)
-			g_free (app->devname);
-		app->devname = dev;
-	}
+	if (app->devname)
+		g_free (app->devname);
+	app->devname = dev;
 
-	g_message ("New connection device is %s", dev ? dev : "empty");
+	g_message ("New connection device is %s (%s)",
+		   dev ? dev : "empty",
+		   changed ? "changed" : "not changed");
 
-	return dev;
-}
-
-static gboolean
-poll_listener (PhonemgrListener *listener)
-{
-	phonemgr_listener_poll (listener);
-	return TRUE;
+	return changed;
 }
 
 static gboolean
@@ -65,13 +86,26 @@ attempt_reconnect (MyApp *app)
 	if (gconf_client_get_bool (app->client,
 				CONFBASE"/auto_retry", NULL) &&
 			! phonemgr_listener_connected (app->listener) &&
-			! app->connecting) {
+			! app->connecting &&
+			app->status == PHONEMGR_LISTENER_IDLE) {
 		g_message ("Auto-retrying the connection");
 		reconnect_phone (app);
 	}
 	return TRUE;
 }
 
+static gboolean
+sync_clock (MyApp *app)
+{
+	if (gconf_client_get_bool (app->client,
+				CONFBASE"/sync_clock", NULL)) {
+		g_message ("Syncing phone clock");
+		phonemgr_listener_set_time (app->listener,
+					    time(NULL));
+	}
+	return TRUE;
+}
+
 static gpointer
 connect_phone_thread (gpointer data)
 {
@@ -89,9 +123,6 @@ connect_phone_thread (gpointer data)
 			/* translators: the '%s' will be substituted with '/dev/ttyS0'
 			   or similar */
 			g_message (_("Connected to device on %s"), app->devname);
-			app->pollsource = g_timeout_add (POLL_TIMEOUT,
-				(GSourceFunc)poll_listener,
-				(gpointer) app->listener);
 		} else {
 			/* the ERROR signal will have been emitted, so we don't
 			   bother changing the icon ourselves at this point */
@@ -120,14 +151,15 @@ static void
 connect_phone (MyApp *app)
 {
 	g_mutex_lock (app->connecting_mutex);
-	if (! phonemgr_listener_connected (app->listener) &&
-			! app->connecting) {
+	if (phonemgr_listener_connected (app->listener) == FALSE
+	    && app->connecting == FALSE
+	    && app->status == PHONEMGR_LISTENER_IDLE) {
 		app->connecting = TRUE;
 		g_mutex_unlock (app->connecting_mutex);
 		/* we're neither connected, nor connecting */
-		app->connecting_thread = g_thread_create (
-				connect_phone_thread, (gpointer) app,
-				FALSE, NULL);
+		app->connecting_thread = g_thread_create (connect_phone_thread,
+							  (gpointer) app,
+							  FALSE, NULL);
 	} else {
 		g_message ("Can't connect twice");
 		g_mutex_unlock (app->connecting_mutex);
@@ -147,12 +179,11 @@ on_status (PhonemgrListener *listener, int status, MyApp *app)
 	g_message ("Status %d", status);
 	gdk_threads_enter ();
 
+	app->status = status;
+
 	switch (status) {
 		case PHONEMGR_LISTENER_IDLE:
 			g_message ("Disconnect complete");
-			/* received when disconnected */
-			g_source_remove (app->pollsource);
-			app->pollsource = 0;
 			set_icon_state (app);
 			if (app->reconnect)
 				g_idle_add ((GSourceFunc)idle_connect_phone,
@@ -163,9 +194,13 @@ on_status (PhonemgrListener *listener, int status, MyApp *app)
 			break;
 		case PHONEMGR_LISTENER_CONNECTED:
 			g_message ("Serial port connected");
+			phonemgr_object_emit_number_batteries_changed (app->object, 1);
+			sync_clock (app);
 			break;
 		case PHONEMGR_LISTENER_DISCONNECTING:
 			g_message ("Closing serial port connection");
+			phonemgr_object_emit_number_batteries_changed (app->object, 0);
+			phonemgr_object_emit_network_registration_changed (app->object, 0, 0, 0, 0);
 			break;
 		case PHONEMGR_LISTENER_ERROR:
 			set_icon_state (app);
@@ -179,8 +214,8 @@ on_status (PhonemgrListener *listener, int status, MyApp *app)
 }
 
 static void
-on_message (PhonemgrListener *listener, gchar *sender,
-		GTime timestamp, gchar *message, MyApp *app)
+on_message (PhonemgrListener *listener, char *sender,
+		time_t timestamp, char *message, MyApp *app)
 {
 	Message *msg;
 
@@ -193,8 +228,6 @@ on_message (PhonemgrListener *listener, gchar *sender,
 	msg->message = g_strdup (message);
 
 	g_mutex_lock (app->message_mutex);
-	if (app->messages == NULL)
-		enable_flasher (app);
 	app->messages = g_list_append (app->messages, (gpointer) msg);
 	g_mutex_unlock (app->message_mutex);
 
@@ -202,10 +235,26 @@ on_message (PhonemgrListener *listener, gchar *sender,
 	play_alert (app);
 }
 
+static void
+on_battery (PhonemgrListener *listener, int percent, gboolean on_ac, MyApp *app)
+{
+	phonemgr_object_emit_battery_state_changed (app->object, 0, percent, on_ac);
+}
+
+static void
+on_network (PhonemgrListener *listener, int mcc, int mnc, int lac, int cid, MyApp *app)
+{
+	phonemgr_object_emit_network_registration_changed (app->object, mcc, mnc, lac, cid);
+}
+
 void
 reconnect_phone (MyApp *app)
 {
-	if (phonemgr_listener_connected (app->listener)) {
+	gboolean changed;
+
+	changed = set_connection_device (app);
+
+	if (changed && phonemgr_listener_connected (app->listener) != FALSE) {
 		g_message ("Disconnecting...");
 		app->reconnect = TRUE;
 		app->disconnecting_thread =
@@ -222,8 +271,6 @@ free_connection (MyApp *app)
 	g_source_remove (app->reconnector);
 	if (phonemgr_listener_connected (app->listener))
 		phonemgr_listener_disconnect (app->listener);
-	if (app->pollsource)
-		g_source_remove (app->pollsource);
 	g_mutex_free (app->connecting_mutex);
 	g_mutex_free (app->message_mutex);
 }
@@ -239,17 +286,6 @@ disconnect_signal_handlers (MyApp *app)
 			app->message_cb);
 }
 
-#ifdef TESTING
-static gboolean
-send_test_message (MyApp *app)
-{
-	g_message ("Sending test message");
-	on_message (app->listener, "1234567", 0, "test message", app);
-	on_message (app->listener, "1234567", 0, "test message 2", app);
-	return FALSE;
-}
-#endif
-
 void
 initialise_connection (MyApp *app)
 {
@@ -257,13 +293,15 @@ initialise_connection (MyApp *app)
 			G_CALLBACK (on_status), (gpointer) app);
 	app->message_cb = g_signal_connect (G_OBJECT (app->listener), "message",
 			G_CALLBACK (on_message), (gpointer) app);
-	app->reconnector = g_timeout_add (20*1000,
-		(GSourceFunc)attempt_reconnect, (gpointer)app);
+	app->battery_cb = g_signal_connect (G_OBJECT (app->listener), "battery",
+						G_CALLBACK (on_battery), (gpointer) app);
+	app->network_cb = g_signal_connect (G_OBJECT (app->listener), "network",
+					    G_CALLBACK (on_network), app);
+	app->reconnector = g_timeout_add_seconds (20, (GSourceFunc)attempt_reconnect,
+						  (gpointer)app);
+
 	gdk_threads_init ();
 	app->connecting_mutex = g_mutex_new ();
 	app->message_mutex = g_mutex_new ();
-#ifdef TESTING
-	g_timeout_add (1000*15, (GSourceFunc) send_test_message,
-	 (gpointer) app);
-#endif
 }
+
