@@ -26,6 +26,7 @@
 #include <glib/gi18n.h>
 
 #include "phonemgr-object.h"
+#include "phonemgr-utils.h"
 #include "app.h"
 
 static char *bdaddr_ident = NULL;
@@ -61,7 +62,6 @@ main (int argc, char **argv)
 	g_option_context_set_translation_domain(context, GETTEXT_PACKAGE);
 
 	g_option_context_add_group (context, gtk_get_option_group (TRUE));
-	g_option_context_add_group (context, gst_init_get_option_group ());
 	if (g_option_context_parse (context, &argc, &argv, &error) == FALSE) {
 		g_error ("Couldn't parse options: %s", error->message);
 		g_error_free (error);
@@ -109,8 +109,6 @@ main (int argc, char **argv)
 	g_object_unref (app->listener);
 	g_object_unref (app->client);
 	g_object_unref (app->object);
-	if (app->playbin)
-		g_object_unref (app->playbin);
 
 	return 0;
 }
